@@ -1,8 +1,9 @@
-from functools import partial
-
 import magma as m
 
 from fault.sva import SVAProperty
+from fault.expression import Expression, UnaryOp, BinaryOp
+from fault.infix import Infix
+from fault.assert_utils import add_compile_guards
 
 
 class Property:
@@ -19,17 +20,6 @@ class Property:
         return self
 
 
-class Infix:
-    def __init__(self, func):
-        self.func = func
-
-    def __or__(self, other):
-        return self.func(other)
-
-    def __ror__(self, other):
-        return Infix(partial(self.func, other))
-
-
 class Implies(Property):
     op_str = "|->"
 
@@ -164,8 +154,8 @@ class _Compiler:
         return f"[{start}:{stop}]"
 
     def _compile(self, value):
-        if isinstance(value, Not):
-            return f"! {self._compile(value.arg)}"
+        if isinstance(value, PropertyUnaryOp):
+            return f"{value.op_str} {self._compile(value.arg)}"
         # TODO: Refactor getitem properties to share code
         if isinstance(value, Delay):
             result = ""
@@ -211,7 +201,8 @@ class _Compiler:
             for arg in value.args:
                 if isinstance(arg, str):
                     property_str += f" {arg} "
-                elif isinstance(arg, (SVAProperty, Sequence, FunctionCall)):
+                elif isinstance(arg, (SVAProperty, Sequence, FunctionCall,
+                                      PropertyUnaryOp, Expression)):
                     property_str += f" {self._compile(arg)} "
                 else:
                     key = f"x{len(self.format_args)}"
@@ -232,6 +223,21 @@ class _Compiler:
             return f"{{{{{contents}}}}}"
         if isinstance(value, int):
             return str(value)
+        if isinstance(value, BinaryOp):
+            left = self._compile(value.left)
+            right = self._compile(value.right)
+            op = value.op_str
+            if op == "==":
+                # Use strict eq
+                op = "==="
+            elif op == "!=":
+                # Use strict neq
+                op = "!=="
+            return f"({left}) {op} ({right})"
+        if isinstance(value, UnaryOp):
+            operand = self._compile(value.operand)
+            op = value.op_str
+            return f"{op} ({operand})"
         raise NotImplementedError(type(value))
 
     def compile(self, prop):
@@ -239,8 +245,8 @@ class _Compiler:
         return compiled
 
 
-def assert_(prop, on, disable_iff=None, compile_guard=None, name=None,
-            inline_wire_prefix="_FAULT_ASSERT_WIRE_"):
+def _make_statement(statement, prop, on, disable_iff, compile_guard, name,
+                    inline_wire_prefix):
     format_args = {}
     _compiler = _Compiler(format_args)
     prop = _compiler.compile(prop)
@@ -248,23 +254,34 @@ def assert_(prop, on, disable_iff=None, compile_guard=None, name=None,
     if disable_iff is not None:
         disable_str = f" disable iff ({_compiler.compile(disable_iff)})"
     event_str = on.compile(format_args)
-    prop_str = f"assert property ({event_str}{disable_str} {prop});"
+    prop_str = f"{statement} property ({event_str}{disable_str} {prop});"
     if name is not None:
-        if not isinstance(compile_guard, str):
+        if not isinstance(name, str):
             raise TypeError("Expected string for name")
         prop_str = f"{name}: {prop_str}"
-    if compile_guard is not None:
-        if not isinstance(compile_guard, str):
-            raise TypeError("Expected string for compile_guard")
-        prop_str = f"""\
-`ifdef {compile_guard}
-    {prop_str}
-`endif
-"""
+    prop_str = add_compile_guards(compile_guard, prop_str)
     m.inline_verilog(prop_str, inline_wire_prefix=inline_wire_prefix,
                      **format_args)
 
 
+def assert_(prop, on, disable_iff=None, compile_guard=None, name=None,
+            inline_wire_prefix="_FAULT_ASSERT_WIRE_"):
+    _make_statement("assert", prop, on, disable_iff, compile_guard, name,
+                    inline_wire_prefix)
+
+
+def cover(prop, on, disable_iff=None, compile_guard=None, name=None,
+          inline_wire_prefix="_FAULT_COVER_WIRE_"):
+    _make_statement("cover", prop, on, disable_iff, compile_guard, name,
+                    inline_wire_prefix)
+
+
+def assume(prop, on, disable_iff=None, compile_guard=None, name=None,
+           inline_wire_prefix="_FAULT_ASSUME_WIRE_"):
+    _make_statement("assume", prop, on, disable_iff, compile_guard, name,
+                    inline_wire_prefix)
+
+
 class Sequence:
     def __init__(self, prop):
         if not isinstance(prop, (Property, SVAProperty)):
@@ -277,7 +294,7 @@ def sequence(prop):
     return Sequence(prop)
 
 
-class FunctionCall:
+class FunctionCall(Expression):
     def __init__(self, func, args):
         self.func = func
         self.args = args
@@ -301,7 +318,13 @@ fell = Function("$fell")
 stable = Function("$stable")
 
 
-class Not(Property):
+class PropertyUnaryOp(Expression):
+    pass
+
+
+class Not(PropertyUnaryOp):
+    op_str = "!"
+
     def __init__(self, arg):
         self.arg = arg
 
