@@ -75,6 +75,11 @@ def error_detected(text, err_str):
         raise Exception(f'Invalid err_str: {err_str}.')
 
 
+def _throw(args, cwd, env, disp_type, err_str,
+           chk_ret_code, shell, use_fault_cfg):
+    raise AssertionError()
+
+
 def subprocess_run(args, cwd=None, env=None, disp_type='on_error', err_str=None,
                    chk_ret_code=True, shell=False, use_fault_cfg=True):
     # "Deluxe" version of subprocess.run that can display STDOUT lines as they
@@ -160,7 +165,13 @@ def subprocess_run(args, cwd=None, env=None, disp_type='on_error', err_str=None,
         print(RED + BRIGHT + f'Found {len(err_msg)} error(s):' + RESET_ALL)
         for k, e in enumerate(err_msg):
             print(RED + BRIGHT + f'{k+1}) {e}' + RESET_ALL)
-        raise AssertionError
+
+        # pytest dumps the source of functions that raise exceptions.
+        # However the source of this function is not relevant to the
+        # exception being raised.  Hence by wrapping `raise AssertionError`
+        # in a function we avoid cluttering the output.
+        _throw(args, cwd, env, disp_type, err_str,
+               chk_ret_code, shell, use_fault_cfg)
 
     # if there were no errors, then return directly
     return CompletedProcess(args=args, returncode=p.returncode,
