@@ -17,12 +17,14 @@ import fault
 import fault.expression as expression
 from fault.ms_types import RealType
 import os
+import pysv
 from numbers import Number
 
 
 src_tpl = """\
 {timescale}
 module {top_module};
+{imports}
 {declarations}
 {assigns}
 {clock_drivers}
@@ -35,6 +37,7 @@ module {top_module};
 
     initial begin
 {initial_body}
+{exit_code}
         #20 $finish;
     end
 
@@ -136,7 +139,7 @@ class SystemVerilogTarget(VerilogTarget):
 
         dump_waveforms: Enable tracing of internal values
 
-        waveform_type: 'vcd', 'vpd', 'fsdb'.  Default for ncsim is 'vcd'. Default
+        waveform_type: 'vcd', 'vpd', 'fsdb'. Default for ncsim is 'vcd'. Default
                        for vcs is 'vpd'.
 
         waveform_file: name of file to dump waveforms (default is
@@ -223,6 +226,7 @@ class SystemVerilogTarget(VerilogTarget):
                           DeprecationWarning)
             self.dump_waveforms = dump_vcd
         self.no_warning = no_warning
+        self.imports = set()
         self.declarations = {}  # dictionary keyed by signal name
         self.assigns = {}  # dictionary keyed by LHS name
         self.sim_env = sim_env if sim_env is not None else {}
@@ -301,6 +305,8 @@ class SystemVerilogTarget(VerilogTarget):
                 name = verilog_name(port[-1].name, self.disable_ndarray)
         elif isinstance(port, fault.WrappedVerilogInternalPort):
             name = f"dut.{port.path}"
+        elif isinstance(port, actions.Var):
+            name = port.name
         else:
             name = verilog_name(port.name, self.disable_ndarray)
         return name
@@ -313,12 +319,16 @@ class SystemVerilogTarget(VerilogTarget):
                 self.disable_ndarray
             )
             return f"dut.{path}"
-        return f"{value.port.name}"
+        return f"{verilog_name(value.port.name, self.disable_ndarray)}"
 
     def make_var(self, i, action):
         if isinstance(action._type, AbstractBitVectorMeta):
             self.add_decl(f'reg [{action._type.size - 1}:0]', action.name)
             return []
+        elif isinstance(action._type, type):
+            self.add_decl(f"{action._type.__name__}", action.name)
+            return []
+
         raise NotImplementedError(action._type)
 
     def make_file_scan_format(self, i, action):
@@ -348,7 +358,7 @@ class SystemVerilogTarget(VerilogTarget):
             new_value = f"{value.file.name_without_ext}_in"
             value = new_value
         elif isinstance(value, expression.Expression):
-            value = f"({self.compile_expression(value)})"
+            value = f"{self.compile_expression(value)}"
         return value
 
     def compile_expression(self, value):
@@ -362,11 +372,11 @@ class SystemVerilogTarget(VerilogTarget):
             elif op == "!=":
                 # Use strict neq
                 op = "!=="
-            return f"({left}) {op} ({right})"
+            return f"({left} {op} {right})"
         elif isinstance(value, expression.UnaryOp):
             operand = self.compile_expression(value.operand)
             op = value.op_str
-            return f"{op} ({operand})"
+            return f"({op} {operand})"
         elif isinstance(value, PortWrapper):
             path = value.select_path.system_verilog_path(self.disable_ndarray)
             return f"dut.{path}"
@@ -374,6 +384,17 @@ class SystemVerilogTarget(VerilogTarget):
             return self.process_peek(value)
         elif isinstance(value, actions.Var):
             value = value.name
+        elif isinstance(value, expression.FunctionCall):
+            args = ", ".join(self.compile_expression(x) for x in value.args)
+            func_str = value.func_str
+            if not isinstance(value, expression.Integer):
+                # special casting syntax
+                func_str = f"${func_str}"
+            return f"{func_str}({args})"
+        elif isinstance(value, int):
+            return str(value)
+        elif isinstance(value, expression.CallExpression):
+            return value.str(True, self.compile_expression)
         return value
 
     def make_poke(self, i, action):
@@ -382,7 +403,12 @@ class SystemVerilogTarget(VerilogTarget):
         value = self.process_value(action.port, action.value)
         # Build up the poke action, including delay
         retval = []
-        retval += [f'{name} <= {value};']
+        # class instance can only be created via blocking assignment
+        if isinstance(action.port, actions.Var) and isinstance(action.port._type, type):
+            assign = "="
+        else:
+            assign = "<="
+        retval += [f'{name} {assign} {value};']
         if action.delay is not None:
             retval += [f'#({action.delay}*1s);']
         return retval
@@ -413,6 +439,21 @@ class SystemVerilogTarget(VerilogTarget):
         self.add_decl('integer', action.loop_var, exist_ok=True)
         return super().make_loop(i, action)
 
+    def make_join(self, i, action):
+        code = ["fork"]
+        for p in action.processes:
+            code += self.make_block(i, None, None, p.actions, p.name)
+        # join
+        if action.join_type == actions.JoinType.Default:
+            code += ["join"]
+        elif action.join_type == actions.JoinType.None_:
+            code += ["join_none"]
+        elif action.join_type == actions.JoinType.Any:
+            code += ["join_any"]
+        else:
+            raise ValueError(f"Unexpected joint_type: {action.join_type}")
+        return code
+
     def make_file_open(self, i, action):
         # make sure the file mode is supported
         if not is_valid_file_mode(action.file.mode):
@@ -490,6 +531,10 @@ class SystemVerilogTarget(VerilogTarget):
         else:
             return [f'if (!({expr_str})) $error("{expr_str} failed");']
 
+    def make_call(self, i, action: actions.Call):
+        self.imports.add("pysv")
+        return super().make_call(i, action)
+
     def make_expect(self, i, action):
         # don't do anything if any value is OK
         if value_utils.is_any(action.value):
@@ -585,6 +630,12 @@ class SystemVerilogTarget(VerilogTarget):
                 port_list.extend(result)
         elif issubclass(type_, m.Tuple):
             for k, t in zip(type_.keys(), type_.types()):
+                try:
+                    int(k)
+                    # python/coreir don't allow pure integer names
+                    k = f"_{k}"
+                except ValueError:
+                    pass
                 result = self.generate_port_code(
                     name + "_" + str(k), t, power_args
                 )
@@ -655,6 +706,9 @@ class SystemVerilogTarget(VerilogTarget):
             port_list.extend(result)
         port_list = f',\n{2*self.TAB}'.join(port_list)
 
+        if len(self.pysv_funcs) > 0:
+            self.imports.add("pysv")
+
         # build up the body of the initial block
         initial_body = []
 
@@ -691,6 +745,10 @@ class SystemVerilogTarget(VerilogTarget):
         initial_body = [f'{2*self.TAB}{elem}' for elem in initial_body]
         initial_body = '\n'.join(initial_body)
 
+        # format imports
+        imports = [f"{self.TAB}import {name}::*;" for name in self.imports]
+        imports = "\n".join(imports)
+
         # format declarations
         declarations = [f'{self.TAB}{type_} {name};'
                         for type_, name in self.declarations.values()]
@@ -706,9 +764,16 @@ class SystemVerilogTarget(VerilogTarget):
 
         clock_drivers = self.TAB + "\n{self.TAB}".join(self.clock_drivers)
 
+        # exit code
+        if len(self.pysv_funcs) > 0:
+            exit_code = f"{self.TAB}pysv_finalize();\n"
+        else:
+            exit_code = ""
+
         # fill out values in the testbench template
         src = src_tpl.format(
             timescale=timescale,
+            imports=imports,
             declarations=declarations,
             clock_drivers=clock_drivers,
             assigns=assigns,
@@ -716,39 +781,50 @@ class SystemVerilogTarget(VerilogTarget):
             port_list=port_list,
             param_list=param_list,
             circuit_name=self.circuit_name,
+            exit_code=exit_code,
             top_module=self.top_module
         )
 
         # return the string representing the system-verilog testbench
         return src
 
-    def run(self, actions, power_args=None):
+    def generate_test_bench(self, actions, power_args=None):
+        if self.ext_test_bench:
+            raise Exception(
+                "Cannot generate_test_bench when using external test bench")
         # set defaults
         power_args = power_args if power_args is not None else {}
+        return self.write_test_bench(actions=actions, power_args=power_args)
 
+    def run(self, actions, power_args=None):
         # assemble list of sources files
         vlog_srcs = []
         if not self.ext_test_bench:
-            tb_file = self.write_test_bench(actions=actions,
-                                            power_args=power_args)
+            tb_file = self.generate_test_bench(actions, power_args)
             vlog_srcs += [tb_file]
         if not self.ext_model_file:
             vlog_srcs += [self.verilog_file]
         vlog_srcs += self.include_verilog_libraries
 
+        lib_path = ""
+        pkg_file = ""
+        if len(self.pysv_funcs) > 0:
+            lib_path = pysv.compile_lib(self.pysv_funcs, self.directory)
+            pkg_file = os.path.join(self.directory, "pysv_pkg.sv")
+            pysv.generate_sv_binding(self.pysv_funcs, filename=pkg_file)
+            assert os.path.exists(pkg_file)
+            vlog_srcs = [os.path.relpath(pkg_file, self.directory)] + vlog_srcs
+
         # generate simulator commands
-        if self.simulator == 'ncsim' or self.simulator == "incisive":
+        if self.simulator in {'ncsim', "incisive", "xcelium"}:
             # Compile and run simulation
             cmd_file = self.write_cadence_tcl()
-            sim_cmd = self.ncsim_cmd(sources=vlog_srcs, cmd_file=cmd_file)
-            sim_err_str = None
-            # Skip "bin_cmd"
-            bin_cmd = None
-            bin_err_str = None
-        elif self.simulator == 'xcelium':
-            # Compile and run simulation
-            cmd_file = self.write_cadence_tcl()
-            sim_cmd = self.xcelium_cmd(sources=vlog_srcs, cmd_file=cmd_file)
+            if self.simulator == "xcelium":
+                sim_cmd = self.xcelium_cmd(sources=vlog_srcs, cmd_file=cmd_file)
+            else:
+                sim_cmd = self.ncsim_cmd(sources=vlog_srcs, cmd_file=cmd_file)
+            if lib_path:
+                sim_cmd += ["-sv_lib", os.path.relpath(lib_path, self.directory)]
             sim_err_str = None
             # Skip "bin_cmd"
             bin_cmd = None
@@ -765,6 +841,8 @@ class SystemVerilogTarget(VerilogTarget):
             # Compile simulation
             # TODO: what error strings are expected at this stage?
             sim_cmd, bin_file = self.vcs_cmd(sources=vlog_srcs)
+            if lib_path:
+                sim_cmd += ["-sv_lib", os.path.relpath(lib_path, self.directory)]
             sim_err_str = None
             # Run simulation
             bin_cmd = [bin_file]
@@ -848,6 +926,7 @@ class SystemVerilogTarget(VerilogTarget):
             tcl_cmds += [f'database -open -vcd vcddb -into {self.waveform_file} -default -timescale ps']  # noqa
             tcl_cmds += [f'probe -create -all -vcd -depth all']
         tcl_cmds += [f'run {self.num_cycles}ns']
+        tcl_cmds += ['assertion -summary -final']
         tcl_cmds += [f'quit']
 
         # write the command file
@@ -983,7 +1062,7 @@ class SystemVerilogTarget(VerilogTarget):
 
             # coverage flags
             if self.coverage:
-                cmd += ["-coverage", "b", "-covoverwrite"]
+                cmd += ["-coverage", "a", "-covoverwrite"]
 
             # return arg list
             return cmd
@@ -1128,4 +1207,4 @@ class SynchronousSystemVerilogTarget(SystemVerilogTarget):
         )
 
     def make_step(self, i, action):
-        return [f"#{self.clock_step_delay * action.steps}"]
+        return [f"#{self.clock_step_delay * action.steps};"]
